/***************************************************************************************
  *
  * InstallDriver.cpp - Source File for Kernel Mode Driver loading.
  *
  * Copyright(c) 2003, Vito Plantamura. All rights reserved.
  *
  * Note: Portions are based on the source found on the www.sysinternals.com site for the filemon application.
  *
***************************************************************************************/

//
// ### Required Header Files. ###
//

#include <windows.h>
#include <tchar.h>
#include <stdio.h>

#include "InstallDriver.h"

//
// ### Definition for stating the type of action to take for the sys file ###
//

#define SYS_FILE_TO_SYSTEM_DIR		0


/****************************************************************************
*
*    FUNCTION: InstallDriver( IN SC_HANDLE, IN LPCTSTR, IN LPCTSTR)
*
*    PURPOSE: Creates a driver service.
*
****************************************************************************/
BOOL InstallDriver( IN SC_HANDLE SchSCManager, IN LPCTSTR DriverName, IN LPCTSTR ServiceExe )
{
    SC_HANDLE  schService;

    //
    // NOTE: This creates an entry for a standalone driver. If this
    //       is modified for use with a driver that requires a Tag,
    //       Group, and/or Dependencies, it may be necessary to
    //       query the registry for existing driver information
    //       (in order to determine a unique Tag, etc.).
    //

    schService = CreateService( SchSCManager,          // SCManager database
                                DriverName,           // name of service
                                DriverName,           // name to display
                                SERVICE_ALL_ACCESS,    // desired access
                                SERVICE_KERNEL_DRIVER, // service type
                                SERVICE_DEMAND_START,  // start type
                                SERVICE_ERROR_NORMAL,  // error control type
                                ServiceExe,            // service's binary
                                NULL,                  // no load ordering group
                                NULL,                  // no tag identifier
                                NULL,                  // no dependencies
                                NULL,                  // LocalSystem account
                                NULL                   // no password
                                );
    if ( schService == NULL )
        return FALSE;

    CloseServiceHandle( schService );

    return TRUE;
}

/****************************************************************************
*
*    FUNCTION: RemoveDriver( IN SC_HANDLE, IN LPCTSTR)
*
*    PURPOSE: Deletes the driver service.
*
****************************************************************************/
BOOL RemoveDriver( IN SC_HANDLE SchSCManager, IN LPCTSTR DriverName )
{
    SC_HANDLE  schService;
    BOOL       ret;

    schService = OpenService( SchSCManager,
                              DriverName,
                              SERVICE_ALL_ACCESS
                              );

    if ( schService == NULL )
        return FALSE;

    ret = DeleteService( schService );

    CloseServiceHandle( schService );

    return ret;
}

/****************************************************************************
*
*    FUNCTION: StartDriver( IN SC_HANDLE, IN LPCTSTR)
*
*    PURPOSE: Starts the driver service.
*
****************************************************************************/

BOOL StartDriver( IN SC_HANDLE SchSCManager, IN LPCTSTR DriverName )
{
    SC_HANDLE  schService;
    BOOL       ret;

    schService = OpenService( SchSCManager,
                              DriverName,
                              SERVICE_ALL_ACCESS
                              );
    if ( schService == NULL )
        return FALSE;

    ret = StartService( schService, 0, NULL )
       || GetLastError() == ERROR_SERVICE_ALREADY_RUNNING 
	   || GetLastError() == ERROR_SERVICE_DISABLED;

    CloseServiceHandle( schService );

    return ret;
}

BOOL StopDriver( IN SC_HANDLE SchSCManager, IN LPCTSTR DriverName )
{
    SC_HANDLE  schService;
    BOOL       ret;

    schService = OpenService( SchSCManager,
                              DriverName,
                              SERVICE_ALL_ACCESS
                              );
    if ( schService == NULL )
        return FALSE;

	SERVICE_STATUS		ssStatus;
	ret = ::ControlService( schService, SERVICE_CONTROL_STOP, & ssStatus );

    CloseServiceHandle( schService );

    return ret;
}

/****************************************************************************
*
*    FUNCTION: OpenDevice( IN LPCTSTR, HANDLE *)
*
*    PURPOSE: Opens the device and returns a handle if desired.
*
****************************************************************************/
BOOL OpenDevice( IN LPCTSTR DriverName, HANDLE * lphDevice )
{
    TCHAR    completeDeviceName[64];
    HANDLE   hDevice;

    //
    // Create a \\.\XXX device name that CreateFile can use
    //
    // NOTE: We're making an assumption here that the driver
    //       has created a symbolic link using it's own name
    //       (i.e. if the driver has the name "XXX" we assume
    //       that it used IoCreateSymbolicLink to create a
    //       symbolic link "\DosDevices\XXX". Usually, there
    //       is this understanding between related apps/drivers.
    //
    //       An application might also peruse the DEVICEMAP
    //       section of the registry, or use the QueryDosDevice
    //       API to enumerate the existing symbolic links in the
    //       system.
    //

	if( (GetVersion() & 0xFF) >= 5 ) {

		//
		// We reference the global name so that the application can
		// be executed in Terminal Services sessions on Win2K
		//
		wsprintf( completeDeviceName, TEXT("\\\\.\\Global\\%s"), DriverName );

	} else {

		wsprintf( completeDeviceName, TEXT("\\\\.\\%s"), DriverName );
	}
    hDevice = CreateFile( completeDeviceName,
                          GENERIC_READ | GENERIC_WRITE,
                          0,
                          NULL,
                          OPEN_EXISTING,
                          FILE_ATTRIBUTE_NORMAL,
                          NULL
                          );
    if ( hDevice == ((HANDLE)-1) )
        return FALSE;

	// If user wants handle, give it to them.  Otherwise, just close it.
	if ( lphDevice )
		*lphDevice = hDevice;
	else
	    CloseHandle( hDevice );

    return TRUE;
}

/****************************************************************************
*
*    FUNCTION: LoadDeviceDriver( const TCHAR, const TCHAR, HANDLE *)
*
*    PURPOSE: Registers a driver with the system configuration manager 
*	 and then loads it.
*
****************************************************************************/
BOOL LoadDeviceDriver( const TCHAR * Name, const TCHAR * Path, 
					  HANDLE * lphDevice, PDWORD Error )
{
	SC_HANDLE	schSCManager;
	BOOL		okay;

	schSCManager = OpenSCManager( NULL, NULL, SC_MANAGER_ALL_ACCESS );

	// Remove previous instance
	RemoveDriver( schSCManager, Name );

	// Ignore success of installation: it may already be installed.
	InstallDriver( schSCManager, Name, Path );

	// Ignore success of start: it may already be started.
	StartDriver( schSCManager, Name );

	// Do make sure we can open it.
	okay = OpenDevice( Name, lphDevice );
	*Error = GetLastError();
 	CloseServiceHandle( schSCManager );

	return okay;
}

/**************************************************
*
*	FUNCTION: UnloadDeviceDriver
*
**************************************************/
BOOL UnloadDeviceDriver( const TCHAR * Name )
{
	BOOL			okay;

	SC_HANDLE	schSCManager;

	schSCManager = OpenSCManager( NULL, NULL, SC_MANAGER_ALL_ACCESS );

	okay = StopDriver( schSCManager, Name );

	CloseServiceHandle( schSCManager );

	return okay;
}

/**************************************************
*
*	FUNCTION:  StartVersionedDeviceDriver
*
**************************************************/
BOOL StartVersionedDeviceDriver( HANDLE * lphDevice, LPCTSTR pszSysFile, LPCTSTR pszSysName, DWORD dwGetVerIoctl, DWORD dwExpectedVerNum, INSTALLDRV_LOGEVENT pfnLogEvent )
{
	TCHAR			Path[ MAX_PATH ];
	HANDLE			findHandle;
	WIN32_FIND_DATA findData;
	TCHAR			*File;
#if SYS_FILE_TO_SYSTEM_DIR
	TCHAR			systemRoot[ MAX_PATH ];
	static TCHAR	driverPath[ MAX_PATH ];
#endif
	DWORD			error;
	DWORD			nb, versionNumber;

	// initialize the passed parameters

	*lphDevice = NULL;

	// search for the driver file in the current directory

	GetCurrentDirectory( sizeof Path, Path );
	_stprintf( Path+lstrlen(Path), _T("\\%s"), pszSysFile );

	findHandle = FindFirstFile( Path, &findData );
	if( findHandle == INVALID_HANDLE_VALUE ) {

		if( !SearchPath( NULL, pszSysFile, NULL, sizeof(Path), Path, &File ) ) {

			if ( pfnLogEvent != NULL )
				pfnLogEvent( INSTALLDRV_MSG_WASNOTFOUND, EVENTLOG_ERROR_TYPE, pszSysFile, NULL );
			return FALSE;
		}

	} else FindClose( findHandle );

#if SYS_FILE_TO_SYSTEM_DIR
	// copy the driver to <winnt>\system32\drivers so that we
	// can run off of a CD or network drive
	if( !GetEnvironmentVariable( "SYSTEMROOT", systemRoot, sizeof(systemRoot))) {

		//strcpy( ErrorDescription, _T("Could not resolve SYSTEMROOT environment variable.") );
		return FALSE;
	}
	sprintf( driverPath, _T("%s\\system32\\drivers\\%s"), 
					systemRoot, pszSysFile );
#endif

	if( !LoadDeviceDriver( pszSysName,
#if SYS_FILE_TO_SYSTEM_DIR
		driverPath,
#else
		Path,
#endif
		lphDevice, &error ) )  {

#if SYS_FILE_TO_SYSTEM_DIR
		if( !CopyFile( Path, driverPath, FALSE )) {

			//sprintf( ErrorDescription, _T("Unable to copy %s to %s. ")
			//	_T("Make sure that %s is in the current directory."), 
			//	pszSysName, driverPath, pszSysFile );
			return FALSE;
		}
		SetFileAttributes( driverPath, FILE_ATTRIBUTE_NORMAL );
		if( !LoadDeviceDriver( pszSysName, driverPath, lphDevice, &error ) )  {
#endif

//			UnloadDeviceDriver( pszSysName );
//			if( !LoadDeviceDriver( pszSysName, driverPath, lphDevice, &error ) )  {

				char		szerror[ MAX_PATH ];
				::sprintf( szerror, "%i", error );

				if ( pfnLogEvent != NULL )
					pfnLogEvent( INSTALLDRV_MSG_ERRORLOADING, EVENTLOG_ERROR_TYPE, pszSysName, Path, szerror, NULL );

#if SYS_FILE_TO_SYSTEM_DIR
				DeleteFile( driverPath );
#endif
				return FALSE;
//			}
#if SYS_FILE_TO_SYSTEM_DIR
		}
		DeleteFile( driverPath );
#endif
	}

	// Correct driver version?
	if( !DeviceIoControl(	*lphDevice, dwGetVerIoctl,
							NULL, 0, &versionNumber, sizeof(DWORD), &nb, NULL ) ||
			versionNumber != dwExpectedVerNum ) {

		if ( pfnLogEvent != NULL )
			pfnLogEvent( INSTALLDRV_MSG_WRONGVERSION, EVENTLOG_ERROR_TYPE, NULL );
		return FALSE;
	}

	// everything went ok

	return TRUE;
}

/**************************************************
* HELPER FUNCTION
**************************************************/

static BOOL GetDriverOutFileFullPath( LPCTSTR pszSysFile, char* pszOutFullPath ) // --> Note: the Buffer (pszOutFullPath) must be MAX_PATH*2 characters in length.
{
	// Calculate the final file name.

	if ( ::GetCurrentDirectory( MAX_PATH, pszOutFullPath ) == 0 )
		return FALSE;
	char		cLastChr = pszOutFullPath[ ::strlen( pszOutFullPath ) - 1 ];
	if ( cLastChr != '\\' && cLastChr != '/' )
		::strcat( pszOutFullPath, "\\" );
	::strcat( pszOutFullPath, pszSysFile );

	// Return to the Caller.

	return TRUE;
}

/**************************************************
*
*	FUNCTION:  CreateDriverFileFromAppResources
*
**************************************************/
BOOL CreateDriverFileFromAppResources( HINSTANCE hResMod, LPCTSTR pszName, LPCTSTR pszType, LPCTSTR pszSysFile )
{
	// Get the full path of the output file.

	char		szOutFullPath[ MAX_PATH * 2 ];
	if ( ::GetDriverOutFileFullPath( pszSysFile, szOutFullPath ) == FALSE )
		return FALSE;

	// Get a memory copy of the Resource.

	HRSRC		hRes = ::FindResource( hResMod, pszName, pszType );
	if ( hRes == NULL )
		return FALSE;

	HGLOBAL		hgResMem = ::LoadResource( hResMod, hRes );
	if ( hgResMem == NULL )
		return FALSE;

	LPVOID		pRes = ::LockResource( hgResMem );
	if ( pRes == NULL )
		return FALSE;

	DWORD		dwResSize = ::SizeofResource( hResMod, hRes );
	if ( dwResSize == 0 )
		return FALSE;

	// Write the Driver File on disk.

	FILE*			pfDriverFile = ::fopen( szOutFullPath, "wb" );
	if ( pfDriverFile == NULL )
		return FALSE;

	// Write the file.

	BOOL		bExitWithErr = FALSE;

	if ( ::fwrite( pRes, 1, dwResSize, pfDriverFile ) != dwResSize )
		bExitWithErr = TRUE;

	::fclose( pfDriverFile );

	// Exit, eventually, returning error.

	if ( bExitWithErr )
	{
		::DeleteFile( szOutFullPath );
		return FALSE;
	}

	// Return to the Caller.

	return TRUE;
}

/**************************************************
*
*	FUNCTION:  DeleteDriverFile
*
**************************************************/
BOOL DeleteDriverFile( LPCTSTR pszSysFile )
{
	// Get the full path of the output file.

	char		szOutFullPath[ MAX_PATH * 2 ];
	if ( ::GetDriverOutFileFullPath( pszSysFile, szOutFullPath ) == FALSE )
		return FALSE;

	// Try to delete the file on disk.

	::DeleteFile( szOutFullPath );

	// Return to the Caller.

	return TRUE;
}
